// Ansteuern eines Standard-LCD mit 4-Bit-Schnittstelle
// und Abfrage von zwei Pins fr Encoder
// Pin 1.0: RS
// Pin 1.1: E
// Pins 1.4 bis 1.7: D4 bis D7
// Pins 1.2 und 1.3: Eingnge A und B fr Encoder (Schalten gegen Masse)
// Version: 2011-10-18

#include "display.h"
#include "io430.h"
#include <stdbool.h>

static const unsigned char RS = BIT0;
static const unsigned char E = BIT1;
static const unsigned char A = BIT2;
static const unsigned char B = BIT3;

// Daten in den oberen vier Bits
static void sendNibble(unsigned char data)
{
  P1OUT = A | B | data; // Pull-up fr Encoder
  P1OUT |= E;
  waitOneMillisecond();
  P1OUT &= ~E;  
  waitOneMillisecond(); // schneller wre, auf die Rckmeldung des Displays zu warten,
                        // aber so sparen wir eine Leitung
}

static void sendByte(unsigned char data)
{
  sendNibble(data & 0xF0u);
  sendNibble(data << 4);
}

static void sendByteRAM(unsigned char data)
{
  sendNibble(data & 0xF0u | RS);
  sendNibble((data << 4) | RS);
}

void initialize(void)
{
  WDTCTL = WDTPW | WDTHOLD;
  
  BCSCTL1 = CALBC1_1MHZ;
  DCOCTL = CALDCO_1MHZ;

  TACTL = MC_2 | TASSEL_2 | TACLR;

  P1OUT = 0xFF; // auch fr Pull-up der beiden Eingnge
  P1REN = A | B; // Pull-up
  P1DIR = ~(A | B);
  
  waitMilliseconds(50);
  sendNibble(0x30u); // Initialisierungsequenz fr den Warmstart
  waitMilliseconds(50);
  sendNibble(0x30u);
  waitMilliseconds(50);
  sendNibble(0x30u);
  sendNibble(0x20u);
  sendByte(0x20u); // Ansteuerung nur einer Zeile, weil fr unser LCD sonst negative Kontrastspannung ntig
  sendByte(0x01u); // lschen  
  sendByte(0x0Cu); // Display an, Cursor aus, Blinken aus
  sendByte(0x06u); // den Inhalt nicht mit dem Cursor schieben
}

void clear(void)
{
  sendByte(0x01u);
}

void putCharacter(unsigned char row, unsigned char column, char character)
{
  sendByte(0x02u); // "home"
  sendByte(0x80u + 0x40u*row + column); // Adresse
  sendByteRAM((unsigned char)character);
}

void writeString(char s[])
{
  bool done = false;
  for(int i = 0; i < 8; i++)
  {
    if(s[i] == 0)
    {
      done = true;
    }
    if(!done)
    {
      putCharacter(0, i, s[i]);
    }
    else
    {
      putCharacter(0, i, ' ');      
    }
  }
}

void writeNumber(unsigned int a)
{
  if(a == 0)
  {
    writeString("       0");
  }
  else
  {
    for(int i = 0; i < 8; i++) // Vorsicht: Stellen von hinten gezhlt
    {
      if(a > 0)
      {
        putCharacter(0, 7-i, '0' + a % 10);
        a /= 10;
      }
      else
      {
        putCharacter(0, 7-i, ' ');
      }
    }
  }
}

void showCursor(bool visible, bool blinking)
{
  sendByte(0x0Cu | (visible?2u:0u) | (blinking?1u:0u));
}

void shiftCursor(bool right)
{
  sendByte(0x10u | (right?0u:1u));
}

void moveCursor(unsigned char row, unsigned char column)
{
  sendByte(0x02u); // "return home"  
  sendByte(0x80u + column + 0x40u*row);
}

static int increment = 0;    
int getIncrement(void)
{
  waitOneMillisecond(); // um sicherzustellen, dass auch wirklich abgefragt wird
  int inc = increment;
  increment = 0; // ab nun Impulse wieder neu zhlen
  return inc;
}

static unsigned int millisecondsSinceLastChange = 0;
unsigned int getMillisecondsSinceLastChange(void)
{
  waitOneMillisecond();
  return millisecondsSinceLastChange;
}

// Wir mssen in dieser Funktion die Impulse vom Encoder zhlen,
// sonst werden die "verschlafen"!
void waitOneMillisecond(void)
{
  static bool firstCall = true;
  static bool oldA = false;
  if(firstCall)
  {
    firstCall = false;
    oldA = ((P1IN & A) != 0);
  }

  unsigned short tStart = TAR;
  unsigned short tEnd = tStart + 1000;
  // ggf. Zeit zum Arbeiten am Anfang und Ende schaffen
  if(tEnd > 0xFFF0)
  {
    tEnd = 0xFFF0;
  }
  else if(tEnd < 0xF)
  {
    tEnd = 0xF;
  }
  
  bool inputDetected = false;
  
  if(tEnd < tStart) // Hatten wir einen berlauf beim Berechnen von tEnd?
  {
    while(TAR > tEnd)
    {
      // Code-Doppelung, siehe unten. Unschn, aber hier schwer zu vermeiden.
      int a = ((P1IN & A) != 0);
      if(a != oldA)
      {
        inputDetected = true;
        bool b = ((P1IN & B) != 0);
        increment = 1;
        if(a == b)
        {
          increment = -1;
        }
        oldA = a;
      }     
    }
  }
  while(TAR < tEnd)
  {
    // Code-Doppelung, siehe oben.
      int a = ((P1IN & A) != 0);
      if(a != oldA)
      {
        inputDetected = true;
        bool b = ((P1IN & B) != 0);
        increment = 1;
      if(a == b)
      {
        increment = -1;
      }
      oldA = a;
    }
  }
  
  if(inputDetected)
  {
    millisecondsSinceLastChange = 0;
  }
  else
  {
    millisecondsSinceLastChange++;
    
    // berlauf rckgngig machen, falls ntig
    if(millisecondsSinceLastChange == 0)
    {
      millisecondsSinceLastChange--;
    }
  }
}

void waitMilliseconds(unsigned int duration)
{
  for(unsigned int i = 0; i < duration; i++)
  {
    waitOneMillisecond();
  }
}